(*
ocamlopt -inline 9999 -unsafe str.cmxa util.ml data.ml cg.ml wsemlm.ml bfgs.ml bitvec.ml adapt.ml
*)

open Array
open Util
open Data

module C = Complex  (* we use complex #s for solving quartics for beta(2,2) priors on psi *)

(* 
   we assume that all important LR optimization stuff is taken care
   of, and that when necessary, we can call out to Cg.optimize of
   Bfgs.optimize to optimize those parameters, for weighted data;
   thankfully, our datav data type already has a datav_indomain field
   that tells us which data points are in domain and which are out of
   domain.  the main concern of this model is three fold: 

     (1) we need to be able to compute the "psi" parameters,
         essentially the generative parameters for the model,
         under some data set

     (2) be able to compute the "pi" parameters

     (3) using these and the CG/BFGS algorithm to perform
         conditional expectation maximization on the whole
         data set

   see:

     Hal Daum\'e III and Daniel Marcu, "Maximum Entropy Genre
     Adaptation Models."  2004.  In preparation.

   for a fuller description of the algorithm employed.

   one special note is that we assume *sorted* feature vectors for
   efficiency reasons; so be careful when plugging this in to
   arbitrary data.  we also assume our "bernoulli implicit" data
   format.
*)

(* just to save typing *)
type ia = int array
type fa = float array

type jnt = fa * fa  (* zn=gen * zn=non-gen *)
type mrg = fa

(* all of our parameters *)
type psi = { psiI : fa; psiO : fa; psiG : fa }
type phi = { phiI : fa; phiO : fa; phiG : fa }
type params =
    { mutable psi    : psi;
      mutable phi    : phi;
      mutable pi     : float;
      mutable pib    : float;
              fcount : int;
              ccount : int }


(* priors for psi *)
type psiPrior = Beta of float | Gaussian of float

(* copying params, checking differences *)
let copy_params p =
  { psi = { psiI = copy p.psi.psiI ; psiO = copy p.psi.psiO ; psiG = copy p.psi.psiG };
    phi = { phiI = copy p.phi.phiI ; phiO = copy p.phi.phiO ; phiG = copy p.phi.phiG };
    pi  = p.pi;    fcount = p.fcount;
    pib = p.pib;   ccount = p.ccount }

let param_sum_abs_diff p p2 =
  let diff  a1 a2 = if a1 == a2 then 0. else fabs (a1 -. a2) in
  let diffL a1 a2 = fabs (a1 /@ a2) in
  let diff_array f a1 a2 = fold_lefti (fun i a1i sm -> sm +. f a1i a2.(i)) 0. a1 in
    diff_array diffL p.psi.psiI p2.psi.psiI +. 
    diff_array diffL p.psi.psiO p2.psi.psiO +. 
    diff_array diffL p.psi.psiG p2.psi.psiG +. 
    diff_array diff  p.phi.phiI p2.phi.phiI +. 
    diff_array diff  p.phi.phiO p2.phi.phiO +. 
    diff_array diff  p.phi.phiG p2.phi.phiG +. 
    diff p.pi  p2.pi  +.
    diff p.pib p2.pib

(* debugging output *)

let dump_expectations f str a =
  Printf.fprintf f "%s =" str;
  for i = 0 to length a - 1 do Printf.fprintf f " %g" a.(i); done;
  Printf.fprintf f "\n";
  flush f;
  ()

let dump_params f p =
  Printf.fprintf f "---------------------------------------------------------------------\n";
  Printf.fprintf f "pi = %8g\tpib = %8g\n" p.pi p.pib;

  dump_expectations f "\npsiI" p.psi.psiI;
  dump_expectations f "psiO" p.psi.psiO;
  dump_expectations f "psiG" p.psi.psiG;

  dump_expectations f "\nphiI" p.phi.phiI;
  dump_expectations f "phiO" p.phi.phiO;
  dump_expectations f "phiG" p.phi.phiG;

  Printf.fprintf f "\n";
  flush f;
  ()


(* before we get to the optimization parts, we need to be able to
   compute some simple expressions *)

(* this is the log probability of a data point according to a data point *)
let gen_log_prob (x : ia) (ps : fa) = 
  let _F = length ps in
  let _P = length x  in
  let v  = ref one in
  let p  = ref 0 in
    for f = 0 to _F - 1 do
      if !p < _P && x.(!p) < f then failwith "gen_log_prob: oops";

      if !p < _P && x.(!p) == f
      then ( v := !v +.  (ps.(f)) ; incr p )
      else ( v := !v +.  (one -@ ps.(f)) );
    done;
    if !p != _P then failwith "gen_log_prob: oops2";
    bounded_log !v

let lr_class_log_prob_binary x y phi =
  let _P = length x in
  let d = ref phi.(0) in
    for i = 0 to _P - 1 do
      let id = x.(i) in
        if id < length phi then d := !d +. phi.(id);
    done;
  let y' = if y == 1 then -1. else 1. in
    one /@ log (1. +. exp (0. +. y' *. !d))

let lr_class_log_prob _C _F x y (phi : fa) =
  if _C == 1 then lr_class_log_prob_binary x y phi else
    let _P = length x in
    let rec dotEx c =
      let d = ref 0. in
        for i = 0 to _P - 1 do
          let id = x.(i) + c * _F in
            if id < length phi then d := !d +. phi.(id);
        done;
        !d in
    let p = init _C (fun i -> dotEx i) in
    let s = fold_left (+@) zero p in
      p.(y) /@ s

let predict_adapt _C _F y x pi psiG psiN phiG phiN =
(*  let psiRat = ref (log ((1. -. pi) /. pi)) in *)
  let psiRat = ref ((log pi) /@ log (1.-.pi)) in
  let p = ref 0 in
  let _P = length x in
(*  let _F = length psiG in *)

    for f = 0 to length psiG - 1 do
      let xnf = if !p < _P && x.(!p) == f then ( incr p; true ) else false in
        if !p < _P && x.(!p) < f then failwith "predict_adapt: oops";
        if xnf
        then psiRat := !psiRat *@ psiG.(f) /@ psiN.(f)
        else psiRat := (!psiRat -@ (!psiRat *@ psiG.(f))) /@ (one -@ psiN.(f))
    done;
    if !p != _P then failwith "predict_adapt: oops2";
    psiRat := one /@ (one +@ !psiRat);

  let dotEx phi init c =
    let d = ref init in
      for i = 0 to _P - 1 do
        let id = x.(i) + c * _F in
          (* if id < length phi then *) d := !d +. phi.(id);
      done;
      !d in

    if _C == 1 then (  (* binary *)
      let p1g = one /@ log (1. +. exp (0. -. dotEx phiG phiG.(0) 0)) in
      let p1n = one /@ log (1. +. exp (0. -. dotEx phiN phiN.(0) 0)) in

      let p1  = (!psiRat *@ p1n         ) +@ ((one -@ !psiRat) *@ p1g         ) in
      let p0  = (!psiRat *@ (one -@ p1n)) +@ ((one -@ !psiRat) *@ (one -@ p1g)) in
      let p   = exp (p1 /@ (p1 +@ p0)) in
(*      Printf.fprintf stderr "predict>> %s y=%d pi=%g psiRat=%g p1g=%g p1n=%g p=%g\n" (if (exp p1g>0.5) == (exp p1n>0.5) then "*" else " ") y pi (exp !psiRat) (exp p1g) (exp p1n) p; *)
        p, if p > 0.5 then 1 else 0
    ) else ( (* multiclass *)
      let aG = init _C (fun c -> dotEx phiG 0. c) in
      let sG = fold_left (+@) zero aG in
      let _  = iteri (fun c v -> aG.(c) <- v /@ sG) aG in

      let aN = init _C (fun c -> dotEx phiN 0. c) in
      let sN = fold_left (+@) zero aN in
      let _  = iteri (fun c v -> aN.(c) <- v /@ sN) aN in

      let p  = init _C (fun c -> (!psiRat *@ aN.(c) +@ log 0.1) +@ ((one -@ !psiRat) *@ aG.(c) +@ log 0.1)) in
      let s  = fold_left (+@) zero p in
      let mx = ref 0 in
        for c = 1 to _C - 1 do
          if p.(c) > p.(!mx) then mx := c;
        done;
(*        dump_expectations stderr "predict, p" aG; *)
        (p.(!mx) /@ s), !mx
    )


let compute_error_rate _C match_domain y x pi psiG psiN phiG phiN =
  let _N = dat_N x in
  let cr = ref 0 in
  let tt = ref 0 in
    for n = 0 to _N - 1 do
      if match_domain x.datav_indomain.(n) then (
        let py,pred = predict_adapt _C x.datav_numfeat y.(n) x.datav_xi.(n) pi psiG psiN phiG phiN in
          if pred == y.(n) then incr cr;
(*          if (py > 0.5 && y.(n) == 1) || (py < 0.5 && y.(n) == 0)
          then incr cr; *)
          incr tt;
      );
    done;
    float_of_int (!tt - !cr) /. float_of_int !tt


let compute_joints
  (x : datav) (y : ia) (p : params)
  : jnt * jnt =
  let _N = length y in
  let _F = p.fcount  in
  let jointIA = make _N zero in
  let jointIB = make _N zero in
  let jointOA = make _N zero in
  let jointOB = make _N zero in
    for n = 0 to _N - 1 do
      if x.datav_indomain.(n) then (
        jointIA.(n) <- log p.pi +. gen_log_prob x.datav_xi.(n) p.psi.psiG +.
                         lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) p.phi.phiG;
        jointIB.(n) <- log (1.-.p.pi) +. gen_log_prob x.datav_xi.(n) p.psi.psiI +.
                         lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) p.phi.phiI;
      ) else (
        jointOA.(n) <- log p.pib +. gen_log_prob x.datav_xi.(n) p.psi.psiG +.
                         lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) p.phi.phiG;
        jointOB.(n) <- log (1.-.p.pib) +. gen_log_prob x.datav_xi.(n) p.psi.psiO +.
                         lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) p.phi.phiO;
      );
    done;

    (jointIA,jointIB), (jointOA, jointOB)



let compute_marginals  (* produces *log* values *)
  (x : datav) (p : params)
  : mrg * mrg =
  let _N = dat_N x  in
  let _F = p.fcount in
  let mrgI = make _N zero in
  let mrgO = make _N zero in
    for n = 0 to _N - 1 do
      if x.datav_indomain.(n) then
        mrgI.(n) <-
          one /@
            (addLog (log p.pi       +. gen_log_prob x.datav_xi.(n) p.psi.psiG)
                    (log (1.-.p.pi) +. gen_log_prob x.datav_xi.(n) p.psi.psiI))
      else 
        mrgO.(n) <-
            one /@
            (addLog (log p.pib       +. gen_log_prob x.datav_xi.(n) p.psi.psiG)
                    (log (1.-.p.pib) +. gen_log_prob x.datav_xi.(n) p.psi.psiO));
    done;
    mrgI, mrgO



let compute_logPSI
  (x : datav) (psG : fa) (psN : fa) (domain : bool) : fa * fa =
  let _N = dat_N x in
  let _F = length psG in
  let lPa = make _N one in let lPb = make _N one in
    for n = 0 to _N - 1 do
      if x.datav_indomain.(n) == domain then (
        let xi = x.datav_xi.(n) in
        let p  = ref 0 in let _P = length xi in
          for f = 0 to _F - 1 do
            let xnf = if !p < _P && xi.(!p) == f then (incr p; true) else false in
              if !p < _P && xi.(!p) < f then failwith "compute_logPSI: oops";
              if xnf
              then ( lPa.(n) <- lPa.(n) *@         psG.(f)  ; lPb.(n) <- lPb.(n) *@         psN.(f)  )
              else ( lPa.(n) <- lPa.(n) *@ (one -@ psG.(f)) ; lPb.(n) <- lPb.(n) *@ (one -@ psN.(f)) );
          done;
          if !p != _P then failwith "compute_logPSI: oops2";
      ) else (
        lPa.(n) <- zero;
        lPb.(n) <- zero
      );
    done;
    lPa, lPb

let compute_logPSI_all
  (x : datav) (psG : fa) (psI : fa) (psO : fa) : fa * fa * fa =
  let _N = dat_N x in
  let _F = length psG in
  let lPG = make _N one  in 
  let lPI = make _N zero in
  let lPO = make _N zero in
    for n = 0 to _N - 1 do
      let lPN,psN = if x.datav_indomain.(n) then lPI,psI else lPO,psO in
      let xi = x.datav_xi.(n) in
      let p  = ref 0 in let _P = length xi in
        lPN.(n) <- one;
        for f = 0 to _F - 1 do
          let xnf = if !p < _P && xi.(!p) == f then (incr p; true) else false in
            if !p < _P && xi.(!p) < f then failwith "compute_logPSI: oops";
            if xnf
            then ( lPG.(n) <- lPG.(n) *@         psG.(f)  ; lPN.(n) <- lPN.(n) *@         psN.(f)  )
            else ( lPG.(n) <- lPG.(n) *@ (one -@ psG.(f)) ; lPN.(n) <- lPN.(n) *@ (one -@ psN.(f)) );
        done;
        if !p != _P then failwith "compute_logPSI: oops2";
    done;
    lPG, lPI, lPO

let xnf_cache = ref (Bitvec.create 0)

let clear_xnf_cache () = Bitvec.clear !xnf_cache

let fill_xnf_cache x f =
  let _N = dat_N x in
  let xnf arr = 
    let _P = length arr in
    let rec xnf' p = 
      if p >= _P then false
      else if arr.(p) == f then true
      else xnf' (p+1) in xnf' 0 in

    clear_xnf_cache ();
    for n = 0 to _N - 1 do
      if xnf x.datav_xi.(n) then Bitvec.unsafe_turn_on !xnf_cache n;
    done;
    ()

let logPSIwithoutF
  n (x : ia) (ps : fa) (lPSI : float) (f : int)
  : float =
    if lPSI >= 0. then 0. else (
      if Bitvec.unsafe_is_on !xnf_cache n
      then lPSI /@ (ps.(f))
      else lPSI /@ (one -@ ps.(f))
    )


let update_logPSI
  (x : datav) (check_domain : bool -> bool) (lPSI : fa) (f : int) (oldv : float) (newv : float)
  : unit =
  let _N = dat_N x in
  let rat_on = newv /@ oldv in 
    for n = 0 to _N - 1 do
      if check_domain x.datav_indomain.(n) then (
        let v =
          if Bitvec.unsafe_is_on !xnf_cache n
          then lPSI.(n) *@ rat_on
          else 
            (one /@ ((one /@ lPSI.(n)) -@ (oldv /@ lPSI.(n)))) -@
            (one /@ ((one /@ (lPSI.(n) *@ newv)) -@ (oldv /@ (lPSI.(n) *@ newv)))) in

          lPSI.(n) <- bounded_log v;
      );
    done;
    ()

let update_logPSI2
  (x : datav) (check_domain : bool -> bool) (lPSI : fa) (ps : fa) (ex : int -> float) pi pib mi mo exi exo
  : float =
  let _N = dat_N x in
  let _F = length ps in
  let likP = ref zero in let likN = ref zero in
    for n = 0 to _N - 1 do
      lPSI.(n) <- zero;
      if check_domain x.datav_indomain.(n) then (
        let xi = x.datav_xi.(n) in
        let _P = length xi in
        let p  = ref 0 in
        let piN, mN, pN =
          if x.datav_indomain.(n)
          then log pi , mi.(n), ex n
          else log pib, mo.(n), ex n in

          lPSI.(n) <- one;
          for f = 0 to _F - 1 do
            let xnf = if !p < _P && xi.(!p) == f then (incr p; true) else false in
              if !p < _P && xi.(!p) < f then failwith "oops";

              if xnf
              then lPSI.(n) <- lPSI.(n) *@ ps.(f)
              else lPSI.(n) <- lPSI.(n) *@ (one -@ ps.(f));

              if xnf
              then likP := !likP +@ (pN *@ ps.(f))
              else likP := !likP +@ (pN *@ (one -@ ps.(f)));
          done;
          if !p != _P then failwith "update_logPSI2: oops2";

          if lPSI.(n) < -8000. then lPSI.(n) <- -8000.;

          likN := !likN +@ (piN *@ mN *@ lPSI.(n));
          (* Printf.fprintf stderr "{%g*%g*%g=%g} " piN mN lPSI.(n) (piN *@ mN *@ lPSI.(n)); *)
      );
    done;
    Printf.fprintf stderr "(ulP2 -> %g-%g = %g) " !likP !likN (!likP -@ !likN);
    !likP -@ !likN

    
(* one of our main tasks is to compute (log) expectations; these are
   just (log) probabilities and are straightforward to compute using the
   above machinery *)

let compute_expectations (* produces logs *)
  (x : datav) (y : ia) (indomain : bool) (pi : float) 
  (logPSIG : fa) (logPSIN : fa) 
  (phiG : fa) (phiN : fa) (p : params)
  : fa =
  let _N = length y in
  let ep = make _N zero in
    for n = 0 to _N - 1 do
      if x.datav_indomain.(n) == indomain then (
        let lPzG = log pi +. 
                     logPSIG.(n) +. 
                     lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) phiG in
        let lPzN = log (1.-.pi) +. 
                     logPSIN.(n) +. 
                     lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) phiN in
          ep.(n) <- lPzG /@ (addLog lPzG lPzN);
          (*Printf.fprintf stderr "ce: %s %d -> (%g %g)=%g / +(%g %g)=%g = %g\n" 
            (if indomain then "id " else "ood") 
            n 
            logPSIG.(n) 
            (lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) phiG) 
            lPzG 
            logPSIN.(n) 
            (lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) phiN) 
            lPzN 
            ep.(n); *)
      );
    done;
    ep


















(* we need to optimize the pi parameter in the Mega model; this is
   relatively easy, especially when no priors are used *)

let opt_pi x domain m logPSIG logPSIN ep =
  (* m is the marginal, logPSIG and logPSIN are precomputed values for
     logPSI for both general "G" and non-general "N" data; ep are the expectations  *)
  let _N = length m in
  let a  = ref 0. in
  let la = ref zero in let lna = ref zero in
  let sumz = ref zero in
    for n = 0 to _N - 1 do
      if x.datav_indomain.(n) == domain then (
        a    := !a    +. exp (m.(n) *@ logPSIN.(n)) -. exp (m.(n) *@ logPSIG.(n));
        la   := !la   +@ (m.(n) *@ logPSIN.(n));
        lna  := !lna  +@ (m.(n) *@ logPSIG.(n));
        sumz := !sumz +@ ep.(n);
      );
    done;
    for n = 0 to _N - 1 do
      if x.datav_indomain.(n) == domain then (
        la  := !la  -@ (m.(n) *@ logPSIG.(n));
        lna := !lna -@ (m.(n) *@ logPSIN.(n));
      );
    done;

    la  := (log 2.) *@ !la ;
    lna := (log 2.) *@ !lna;

    if fabs !a < 1e-10 then (
      let ret = !sumz /@ (((log 2.) *@ !sumz) -@ one) in
        Printf.fprintf stderr "opt_pi 1A: a=%g la=%g lna=%g sumz=%g ret=%g\n" !a !la !lna !sumz ret;
        ret
    ) else if !a < 0. then (

      if (log 0.5 +@ (one /@ ((log 2.) *@ !lna))) > !sumz /@ !lna then ( (* b/2(-a) > 0 *)
        let b_2na = (log 0.5 +@ (one /@ ((log 2.) *@ !lna))) -@ (!sumz /@ !lna) in
        let c_na  = !sumz /@ !lna in
        let ret1  = b_2na +@ (0.5 *. ((2. *. b_2na) +@ c_na)) in
        let ret2  = b_2na -@ (0.5 *. ((2. *. b_2na) +@ c_na)) in
(*        let ret   = if fabs (ret1 -. log 0.5) < fabs (ret2 -. log 0.5) && ret1 < 0. then ret1 else ret2 in *)
        let ret   = if ret1 < 0. then ret1 else ret2 in
          Printf.fprintf stderr "opt_pi 2A: a=%g lna=%g sumz=%g b_2na=%g c_na=%g ret=%g\n" !a !lna !sumz b_2na c_na ret;
          ret
      ) else ( (* b/2(-a) < 0 *)
        let nb_2na = (!sumz /@ !lna) -@ (log 0.5 +@ (one /@ ((log 2.) *@ !lna))) in
        let c_na  = !sumz /@ !lna in
        let ret   = (0.5 *. ((2. *. nb_2na) +@ c_na)) -@ nb_2na in
          Printf.fprintf stderr "opt_pi 2B: a=%g lna=%g sumz=%g nb_2na=%g c_na=%g ret=%g\n" !a !lna !sumz nb_2na c_na ret;
          ret
      )

    ) else (* !a > 0 *) (

      if (log 0.5 +@ (!sumz /@ !la)) > (one /@ (log 2. *@ !la)) then (
        let nb_2a = (log 0.5 +@ (!sumz /@ !la)) -@ (one /@ (log 2. *@ !la)) in
        let c_a   = !sumz /@ !la in
        let ret1  = nb_2a -@ (0.5 *. ((2. *. nb_2a) -@ c_a)) in
        let ret2  = nb_2a +@ (0.5 *. ((2. *. nb_2a) -@ c_a)) in
        let ret   = if ret1 < 0. then ret1 else ret2 in
(*        let ret   = if fabs (ret1 -. log 0.5) < fabs (ret2 -. log 0.5) && ret1 <= 0. then ret1 else ret2 in *)
          Printf.fprintf stderr "opt_pi 3A: a=%g la=%g sumz=%g nb_2a=%g c_a=%g ret=%g\n" !a !la !sumz nb_2a c_a ret;
          ret
      ) else (
        let b_2a  = (one /@ (log 2. *@ !la)) -@ (log 0.5 +@ (!sumz /@ !la)) in
        let c_a   = !sumz /@ !la in
        let ret   = (0.5 *. ((2. *. b_2a) -@ c_a)) -@ b_2a in
          Printf.fprintf stderr "opt_pi 3B: a=%g la=%g sumz=%g b_2a=%g c_a=%g ret=%g\n" !a !la !sumz b_2a c_a ret;
          ret
      )

    )

(* optimize psi, with a beta(a,a) prior: *)
let mstep_psi_nongen_beta
  (xd : datav) (m : fa) (ps : fa) (lPSI : fa) 
  (pi : float) (ex : fa) (domain : bool) (beta_val : float)
  : unit =
  let _F = length ps in
  let _N = dat_N xd  in
  let tot = ref zero in
    for f = 0 to _F - 1 do ps.(f) <- zero; done;
    for n = 0 to _N - 1 do
      if xd.datav_indomain.(n) == domain then (
        let pG = one -@ ex.(n) in
        let xi = xd.datav_xi.(n) in
        let _P = length xi in
          tot := !tot +@ pG;
          for p = 0 to _P - 1 do
            ps.(xi.(p)) <- ps.(xi.(p)) +@ pG;
          done;
      );
    done;
    if beta_val >= 1. then (
      tot := !tot +@ log (2. *. (beta_val -. 1.));
      for f = 0 to _F - 1 do 
        ps.(f) <- bounded_log ((ps.(f) +@ log (beta_val -. 1.)) /@ !tot); 
      done;
    ) else (
      tot := !tot -@ log (2. *. (1. -. beta_val));
      for f = 0 to _F - 1 do 
        ps.(f) <- bounded_log ((ps.(f) -@ log (1. -. beta_val)) /@ !tot); 
      done;
    );
    let _ = update_logPSI2 xd (fun d -> d == domain) lPSI ps (fun n -> one -@ ex.(n)) (1.-.pi) (1.-.pi) m m ex ex in
    Printf.fprintf stderr "psi%s updated\n" (if domain then "I" else "O");
    ()

let mstep_psi_gen_beta
  (x : datav)
  (mI : fa)  (mO : fa)
  (ps : fa) (lPSIG : fa)
  (pi : float) (pib : float) (exI : fa) (exO : fa) (beta_val : float)
  : unit =
  let _F = length ps in
  let _N = dat_N x in
  let tot = ref zero in
    for f = 0 to _F - 1 do ps.(f) <- zero; done;
    for n = 0 to _N - 1 do
      let pG = if x.datav_indomain.(n) then exI.(n) else exO.(n) in
      let xi = x.datav_xi.(n) in
      let _P = length xi in
        tot := !tot +@ pG;
        for p = 0 to _P - 1 do
          ps.(xi.(p)) <- ps.(xi.(p)) +@ pG;
        done;
    done;
    if beta_val >= 1. then (
      tot := !tot +@ log (2. *. (beta_val -. 1.));
      for f = 0 to _F - 1 do 
        ps.(f) <- bounded_log ((ps.(f) +@ log (beta_val -. 1.)) /@ !tot); 
      done;
    ) else (
      tot := !tot -@ log (2. *. (1. -. beta_val));
      for f = 0 to _F - 1 do 
        ps.(f) <- bounded_log ((ps.(f) -@ log (1. -. beta_val)) /@ !tot); 
      done;
    );
    let _ = update_logPSI2 x (fun _ -> true) lPSIG ps (fun n -> if x.datav_indomain.(n) then exI.(n) else exO.(n)) pi pib mI mO exI exO in
    Printf.fprintf stderr "psiG updated\n";
    ()



(* for gaussian priors *)
(* todo: gaussians need to be log-ified *)
let mstep_psi_nongen_gaussian
  (xd : datav) (m : fa) (ps : fa) (lPSI : fa) (pi : float) (ex : fa) (lambda : float) (domain : bool)
  : unit =
  let _F = length ps in
  let sumEp0 = fold_lefti (fun n ep sm -> 
                             if xd.datav_indomain.(n) == domain 
                             then sm +. (1. -. exp ep)
                             else sm) 0. ex in
  let rec iter inum =
    let dw = ref 0. in
    for f = 0 to _F - 1 do
      let oldps = ps.(f) in
      for ii = 1 to 100 do
        let sm = fold_lefti (fun n mn sm -> 
                               if xd.datav_indomain.(n) == domain
                               then sm +. mn *. (1.-.pi) *. exp (logPSIwithoutF n xd.datav_xi.(n) ps lPSI.(n) f)
                               else sm) 0. m in
        let g = (lambda *. lambda *. (log (1. -. ps.(f)) -. log ps.(f))
                 -. ps.(f) *. sumEp0) /. ps.(f) /. (1. -. ps.(f)) -. sm in
        let h = (lambda *. lambda) /. ps.(f) /. (1. -. ps.(f)) /. ps.(f) /. (1. -. ps.(f)) *.
                  (log (1. -. ps.(f)) -. log ps.(f) -. 1.)
                -. sumEp0 /. ps.(f) /. (1. -. ps.(f)) in
        let temp = exp (-0.1 *. float_of_int ii) in
        let new_psf = bounded (ps.(f) -. temp *. g/.h) in
          update_logPSI xd (fun d -> d == domain) lPSI f ps.(f) new_psf; 
          ps.(f) <- new_psf; 
      done;
      dw := max !dw (fabs (oldps -. ps.(f)));
      flush stderr;
    done;
    Printf.fprintf stderr "iter %d, maxdw = %g \n" inum !dw;
    flush stderr;
    if inum < 100 && !dw > 1e-8 then iter (inum+1)
  in
    iter 1;
    ()

let mstep_psi_gen_gaussian
  (x : datav)
  (mI : fa)  (mO : fa)
  (ps : fa) (lPSI : fa) 
  (pi : float) (pib : float) 
  (exI : fa) (exO : fa) 
  (lambda : float)
  : unit =
  let _F = length ps in
  let _N = dat_N  x in
  let fabs x = if x <= 0. then 0. -. x else x in
  let sumEp0 = 
    fold_lefti (fun n exOn sm ->
                  if x.datav_indomain.(n) 
                  then sm +. exp exI.(n)
                  else sm +. exp exOn) 0. exO in
  let rec iter inum =
    let dw = ref 0. in
    for f = 0 to _F - 1 do
      let oldps = ps.(f) in
      for ii = 1 to 100 do
        let sm =
          fold_lefti (fun n mOn sm ->
                        if x.datav_indomain.(n)
                        then sm +. mI.(n) *. pi  *. exp (logPSIwithoutF n x.datav_xi.(n) ps lPSI.(n) f)
                        else sm +. mOn    *. pib *. exp (logPSIwithoutF n x.datav_xi.(n) ps lPSI.(n) f)
                     ) 0. mO in
        let g = (lambda *. lambda *. (log (1. -. ps.(f)) -. log ps.(f))
                 -. ps.(f) *. sumEp0) /. ps.(f) /. (1. -. ps.(f)) -. sm in
        let h = (lambda *. lambda) /. ps.(f) /. (1. -. ps.(f)) /. ps.(f) /. (1. -. ps.(f)) *.
                  (log (1. -. ps.(f)) -. log ps.(f) -. 1.)
                -. sumEp0 /. ps.(f) /. (1. -. ps.(f)) in
(*        let _ = Printf.fprintf stderr "f=%d ii=%d\tg=%g\th=%g\n" f ii g h in *)
        let temp = exp (-0.1 *. float_of_int ii) in
        let new_psf = bounded (ps.(f) -. temp *. g/.h) in
          update_logPSI x (fun _ -> true) lPSI f ps.(f) new_psf;
          ps.(f) <- new_psf;
      done;
      dw := max !dw (fabs (oldps -. ps.(f)));
    done;
    Printf.fprintf stderr "iter %d, maxdw = %g \n" inum !dw;
    flush stderr;
    if inum < 100 && !dw > 1e-8 then iter (inum+1)
  in
    iter 1;
    ()




(* wrapper for CG/BFGS stuff is here *)

let mstep_phi_nongen
  (x : datav) (y : ia) (lambda : float) (ex : fa) (domain : bool) (multiclass : bool) : fa = 
  let _N   = length y in
  let my_y = init _N (fun i -> if y.(i) == 1 then 1. else -1.) in
  let _    = iteri (fun i exi -> 
                      if x.datav_indomain.(i) == domain
                      then x.datav_weights.(i) <- 0.5 +. exp (subLog one exi)
                      else x.datav_weights.(i) <- 0.1) ex in
(*  let maxW = let v = fold_left max 0. x.datav_weights in if v <= 0. then 0.1 else v in
  let _    = iteri (fun i v -> x.datav_weights.(i) <- v /. maxW) x.datav_weights in *)
(*  let _    = dump_expectations stderr "phi weights" x.datav_weights in *)
    if multiclass
    then Bfgs.optimize 100 false lambda 1e-10 5 x y    empty_data_set [||] empty_data_set [||] None
    else Cg.compute    100 false lambda 1e-10   x my_y empty_data_set [||] empty_data_set [||]

let mstep_phi_gen
  (x : datav) (y : ia) (lambda : float) (exI : fa) (exO : fa) (multiclass : bool) : fa =
  let _N   = dat_N x in 
  let my_y = init _N (fun i -> if y.(i) == 1 then 1. else -1.) in
  let _    = iteri (fun i exo -> 
                      if x.datav_indomain.(i)
                      then x.datav_weights.(i) <- 0.5 +. exp exI.(i)
                      else x.datav_weights.(i) <- 0.5 +. exp exo    ) exO in
(*  let maxW = let v = fold_left max 0. x.datav_weights in if v <= 0. then 0.1 else v in
  let _    = iteri (fun i v -> x.datav_weights.(i) <- v /. maxW) x.datav_weights in *)
(*  let _    = dump_expectations stderr "phi weights" x.datav_weights in *)
    if multiclass
    then Bfgs.optimize 100 false lambda 1e-10 5 x y    empty_data_set [||] empty_data_set [||] None
    else Cg.compute    100 false lambda 1e-10   x my_y empty_data_set [||] empty_data_set [||]







(* CEM stuff follows *)

(* we want to monitor complete and incomplete log likelihood and posterior *)

let compute_complete_log_likelihood lambda x y epI epO lPSII lPSIO lPSIG p mi mo =
  let norm_phi phi si =
    fold_lefti (fun i phii sm ->
                  sm -. 0.5 *. lambda *. phii *. phii) 0. phi in

  let ll = ref one in

  let _N = dat_N x  in
    for n = 0 to _N - 1 do
      let zn    = if x.datav_indomain.(n) then epI.(n) else epO.(n) in
      let pi    = if x.datav_indomain.(n) then p.pi    else p.pib   in
      let lPSIN = if x.datav_indomain.(n) then lPSII   else lPSIO   in
      let phiN  = if x.datav_indomain.(n) then p.phi.phiI else p.phi.phiO in
        ll := !ll
          *@ ((log pi) ^@ zn) 
          *@ ((log (1.-.pi)) ^@ (one -@ zn))
          *@ ((lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) p.phi.phiG) ^@ zn)
          *@ ((lr_class_log_prob p.ccount p.fcount x.datav_xi.(n) y.(n) phiN) ^@ (one -@ zn))
          *@ ((lPSIG.(n)) ^@ zn)
          *@ ((lPSIN.(n)) ^@ (one -@ zn));
    done;
    !ll

let compute_incomplete_log_likelihood x (jntIA,jntIB) (jntOA,jntOB) =
  let llN = ref one in let llG = ref one in
  let _N = dat_N x in
    for n = 0 to _N - 1 do
      if x.datav_indomain.(n) 
      then ( llG := !llG +. jntIA.(n) ; llN := !llN +. jntIB.(n) )
      else ( llG := !llG +. jntOA.(n) ; llN := !llN +. jntOB.(n) );
    done;
    addLog !llN !llG

let compute_log_prior phiSigma2 psiPrior p =
  let norm_phi phi si =
    fold_lefti (fun i phii sm ->
                  sm -. 0.5 /. si *. phii *. phii) one phi in
  let norm_psi_beta p psi =
    fold_lefti (fun i psii sm ->
                  sm +. (p-.1.) *. psii +. (p-.1.) *. (subLog one psii)) one psi in
    norm_phi p.phi.phiI phiSigma2 +. 
    norm_phi p.phi.phiO phiSigma2 +. 
    norm_phi p.phi.phiG phiSigma2 +.
    (match psiPrior with
         Beta bv -> (norm_psi_beta bv p.psi.psiI +.
                     norm_psi_beta bv p.psi.psiO +.
                     norm_psi_beta bv p.psi.psiG)
       | Gaussian l -> failwith "gaussian prior not implemented")
(*
    (match psiPrior with
         Beta00 -> (norm_psi_beta 0. p.psi.psiI +.
                    norm_psi_beta 0. p.psi.psiO +.
                    norm_psi_beta 0. p.psi.psiG)
       | Beta11 -> (norm_psi_beta 1. p.psi.psiI +.
                    norm_psi_beta 1. p.psi.psiO +.
                    norm_psi_beta 1. p.psi.psiG)
       | Beta22 -> (norm_psi_beta 2. p.psi.psiI +.
                    norm_psi_beta 2. p.psi.psiO +.
                    norm_psi_beta 2. p.psi.psiG)
       | Gaussian l -> (norm_phi p.psi.psiI l +.  (* todo: logify this *)
                        norm_phi p.psi.psiO l +.
                        norm_phi p.psi.psiG l))
*)

(* initialization of CEM parameters *)

let initialize_parameters (x : datav) (y : ia) : params =
  let cf = dat_F x in
  let _F = x.datav_numfeat+1 in

  let psiI = make _F (log 0.5) in let psiO = make _F (log 0.5) in let psiG = make _F (log 0.5) in
  let numI = ref 0 in

(*
    for n = 0 to dat_N x - 1 do
      if x.datav_indomain.(n) then incr numI;

      for p = 0 to length x.datav_xi.(n) - 1 do
        let f = x.datav_xi.(n).(p) in
          psiG.(f) <- psiG.(f) +. 1.;
          if x.datav_indomain.(n)
          then psiI.(f) <- psiI.(f) +. 1.
          else psiO.(f) <- psiO.(f) +. 1.;
      done;
    done;

    Printf.fprintf stderr "N=%d numI=%d\n" (dat_N x) !numI;

    for f = 0 to _F - 1 do
      psiG.(f) <- bounded_log (log (psiG.(f) /. float_of_int (2 + dat_N x)));
      psiI.(f) <- bounded_log (log (psiI.(f) /. float_of_int (2 + !numI)));
      psiO.(f) <- bounded_log (log (psiO.(f) /. float_of_int (2 + dat_N x - !numI)));
      psiG.(f) <- log 0.1;
      psiI.(f) <- log 0.1;
      psiO.(f) <- log 0.1;
    done;
*)

  let psi0 = { psiI = psiI ; psiO = psiO; psiG = psiG } in
  let phi0 = { phiI = make cf 0. ; phiO = make cf 0. ; phiG = make cf 0. } in
    xnf_cache := Bitvec.create (dat_N x);
    { psi = psi0 ; phi = phi0 ; pi = 0.5 ; pib = 0.5 ; fcount = _F ; ccount = x.datav_numclass }

(*
    dump_expectations stderr "psiI" p.psi.psiI;
    dump_expectations stderr "psiO" p.psi.psiO;
    dump_expectations stderr "psiG" p.psi.psiG;

    dump_expectations stderr "lPSII" lPSII;
    dump_expectations stderr "lPSIO" lPSIO;
    dump_expectations stderr "lPSIG" lPSIG;
*)

(* perform one iteration of CEM *)

let cem_iteration (x : datav) (y : ia) (p : params) (lPSII : fa) (lPSIO : fa) (lPSIG : fa) (lambda : float) 
  (psiPrior : psiPrior) (multiclass : bool) =
  let _N  = dat_N x in
  let exI   = compute_expectations x y true  p.pi  lPSIG lPSII p.phi.phiG p.phi.phiI p in
  let exO   = compute_expectations x y false p.pib lPSIG lPSIO p.phi.phiG p.phi.phiO p in
(*  let _     = dump_expectations stderr "exI" exI in
  let _     = dump_expectations stderr "exO" exO in
*)

    (* compute marginals, joints *)
  let mrgI,mrgO = compute_marginals x   p in
  let jntI,jntO = compute_joints    x y p in
(*  let _         = dump_expectations stderr "mrgI" mrgI in
  let _         = dump_expectations stderr "mrgO" mrgO in *)

  let cll = compute_complete_log_likelihood lambda x y exI exO lPSII lPSIO lPSIG p mrgI mrgO in
  let ill = compute_incomplete_log_likelihood x jntI jntO in
  let _   = Printf.fprintf stderr "CLL1 = %g\tILL = %g\n" cll ill in


(*  
  let _         = dump_expectations stderr "jntI1" (fst jntI) in
  let _         = dump_expectations stderr "jntI2" (snd jntI) in
  let _         = dump_expectations stderr "jntO1" (fst jntO) in
  let _         = dump_expectations stderr "jntO2" (snd jntO) in
*)

    (* update phis *)
  let phiI' = mstep_phi_nongen x y lambda exI     true  multiclass in
  let phiO' = mstep_phi_nongen x y lambda     exO false multiclass in
  let phiG' = mstep_phi_gen    x y lambda exI exO       multiclass in
  let _     = p.phi <- { phiI = phiI' ; phiO = phiO' ; phiG = phiG' } in

  let cll = compute_complete_log_likelihood lambda x y exI exO lPSII lPSIO lPSIG p mrgI mrgO in
  let ill = compute_incomplete_log_likelihood x jntI jntO in
  let _   = Printf.fprintf stderr "CLL2 = %g\tILL = %g\n" cll ill in

    (* update pis *)
(*  let pi'   = bounded (exp (opt_pi x true  mrgI lPSIG lPSII exI)) in
  let pib'  = bounded (exp (opt_pi x false mrgO lPSIG lPSIO exO)) in  
    Printf.fprintf stderr "pi=%g pib=%g\n" pi' pib';

    p.pi <- pi' ; p.pib <- pib';
*)
  let pi', pib' = 0.5, 0.5 in

    (* update psis *)

  let () = 
    match psiPrior with
        Beta beta_val ->
          ( mstep_psi_nongen_beta x mrgI      p.psi.psiI lPSII pi'      exI     true  beta_val
          ; mstep_psi_nongen_beta x      mrgO p.psi.psiO lPSIO     pib'     exO false beta_val
          ; mstep_psi_gen_beta    x mrgI mrgO p.psi.psiG lPSIG pi' pib' exI exO       beta_val
          )
      | Gaussian lambda ->
          ( mstep_psi_nongen_gaussian x mrgI      p.psi.psiI lPSII pi'      exI     lambda true
          ; mstep_psi_nongen_gaussian x mrgO      p.psi.psiO lPSIO     pib'     exO lambda false
          ; mstep_psi_gen_gaussian    x mrgI mrgO p.psi.psiG lPSIG pi' pib' exI exO lambda       ) in


  let cll = compute_complete_log_likelihood lambda x y exI exO lPSII lPSIO lPSIG p mrgI mrgO in
  let ill = compute_incomplete_log_likelihood x jntI jntO in
  let _   = Printf.fprintf stderr "CLL3 = %g\tILL = %g\n" cll ill in

    flush stderr; 
    ()

(*
let testX =
  { datav_bernoulli = true;
    datav_implicit  = true;
    datav_sampled   = false;
    datav_numclass  = 1;
    datav_numfeat   = 5;
    datav_indomain  = [| true; true; true; true; true; true; false; false; false; false; false; false |];
    datav_weights   = [| 1.; 1.; 1.; 1.; 1.; 1.; 1.; 1.; 1.; 1.; 1.; 1. |];
    datav_xi        = [| [|0;2;3;4|] ; [|0;3;4|] ; [|0;2|] ; [|0;4;5|] ; [|0;5|] ; [|0;5|] ;
                         [|0;1;3;4|] ; [|0;3;4|] ; [|0;1|] ; [|0;4;5|] ; [|0;5|] ; [|0;5|] |];
    datav_xi_c      = [||];
    datav_xv        = [||];
    datav_xv_c      = [||] }

let testY = [| 1;1;1;0;0;0 ; 1;1;1;0;0;0 |]
*)

let filename, lambda, betaval, numiter = 
  match Sys.argv with
      [| exec ; fn ; la ; bv ; ni |] -> fn, float_of_string la, float_of_string bv, int_of_string ni
    | _ -> failwith ("usage: " ^ Sys.argv.(0) ^ " <file> <lambda> <beta-val> <numiter>")

let h = open_in filename
let ((trX,trY), (deX,deY), (teX,teY)) = load_implicit true h int_of_string_err 0 false (fun _ _ _ -> ())
let _ = close_in h
let _ =
  Printf.fprintf stderr "Sorting training data...\n"; flush stderr;
  for n = 0 to dat_N trX-1 do fast_sort compare trX.datav_xi.(n); done;

  Printf.fprintf stderr "Sorting development data...\n"; flush stderr;
  for n = 0 to dat_N deX-1 do fast_sort compare deX.datav_xi.(n); done;

  Printf.fprintf stderr "Sorting testing data...\n"; flush stderr;
  for n = 0 to dat_N teX-1 do fast_sort compare teX.datav_xi.(n); done;

  let nc = fold_left max 0 trY in
    trX.datav_numclass <- nc+1;
    deX.datav_numclass <- nc+1;
    teX.datav_numclass <- nc+1;
  ()

let _ = ( Printf.fprintf stderr "Initializing parameters...\n"; flush stderr )
let p = initialize_parameters trX trY
let _ = ( Printf.fprintf stderr "Initializing log PSIs...\n"; flush stderr )
let lPSII, lPSIO, lPSIG = 
  let lPSIG0, lPSII0, lPSIO0 = compute_logPSI_all trX p.psi.psiG p.psi.psiI p.psi.psiO in
    ref lPSII0, ref lPSIO0, ref lPSIG0

let _ =
  let id x = x in
  for iter = 1 to numiter do
    Printf.fprintf stderr "\n\n================= ITERATION %d =================\n\n" iter;
    let p_old = copy_params p in
      cem_iteration trX trY p !lPSII !lPSIO !lPSIG lambda (Beta betaval) false;
(*      dump_params stderr p; *)
      Printf.fprintf stderr "\n=========== difference = %g ===========\n" (param_sum_abs_diff p p_old);
      let ide  = compute_error_rate p.ccount id  trY trX p.pi  p.psi.psiG p.psi.psiI p.phi.phiG p.phi.phiI in
      let ode  = compute_error_rate p.ccount not trY trX p.pib p.psi.psiG p.psi.psiO p.phi.phiG p.phi.phiO in
      let ide2 = compute_error_rate p.ccount id  deY deX p.pi  p.psi.psiG p.psi.psiI p.phi.phiG p.phi.phiI in
      let ode2 = compute_error_rate p.ccount not deY deX p.pib p.psi.psiG p.psi.psiO p.phi.phiG p.phi.phiO in
      let ide3 = compute_error_rate p.ccount id  teY teX p.pi  p.psi.psiG p.psi.psiI p.phi.phiG p.phi.phiI in
      let ode3 = compute_error_rate p.ccount not teY teX p.pib p.psi.psiG p.psi.psiO p.phi.phiG p.phi.phiO in
      Printf.fprintf stderr "train: id error = %g\tood error = %g\ttot error = %g\n" ide  ode  (ide  +. ode );
      Printf.fprintf stderr "dev  : id error = %g\tood error = %g\ttot error = %g\n" ide2 ode2 (ide2 +. ode2);
      Printf.fprintf stderr "test : id error = %g\tood error = %g\ttot error = %g\n" ide3 ode3 (ide3 +. ode3);
      flush stderr;
  done;
  ()


(*
let h = open_in "x5"
let ((testX, testY),_,(testsX, testsY)) = load_implicit true h int_of_string_err 0 false (fun _ _ _ -> ())

let _ =  (* sort the datav_xi-s *)
  let _N = dat_N testX in
    for n = 0 to _N - 1 do
      fast_sort compare testX.datav_xi.(n);
    done;
  let _N = dat_N testsX in
    for n = 0 to _N - 1 do
      fast_sort compare testsX.datav_xi.(n);
    done;
    ()


let p = initialize_parameters testX testY
let lPSII0, lPSIO0, lPSIG0 = initialize_logPSIs testX p
let lPSII, lPSIO, lPSIG = ref lPSII0, ref lPSIO0, ref lPSIG0


let _ =
  for iter = 1 to 100 do
    let p_old = copy_params p in
      Printf.fprintf stderr "\n\n================= ITERATION %d =================\n\n" iter;
      cem_iteration testX testY p !lPSII !lPSIO !lPSIG 1. (Beta 2.) false;
      dump_params stderr p;
      Printf.fprintf stderr "\n=========== difference = %g ===========\n" (param_sum_abs_diff p p_old);
      let ide  = compute_error_rate id  testY  testX  p.pi  p.psi.psiG p.psi.psiI p.phi.phiG p.phi.phiI in
      let ode  = compute_error_rate (fun x -> not x) testY  testX  p.pib p.psi.psiG p.psi.psiO p.phi.phiG p.phi.phiO in
      let ide2 = compute_error_rate id  testsY testsX p.pi  p.psi.psiG p.psi.psiI p.phi.phiG p.phi.phiI in
      let ode2 = compute_error_rate (fun x -> not x) testsY testsX p.pib p.psi.psiG p.psi.psiO p.phi.phiG p.phi.phiO in
      Printf.fprintf stderr "train: id error = %g\tood error = %g\ttot error = %g\n" ide  ode  (ide  +. ode );
      Printf.fprintf stderr "test : id error = %g\tood error = %g\ttot error = %g\n" ide2 ode2 (ide2 +. ode2);
      flush stderr;
  done;
  ()

*)
