open Array
open Util
open Data

type memory =
    { mutable mem_filled : bool;
      mutable mem_posn   : int;
      mem_darray : bigarr array;
      mem_uarray : bigarr array;
      mem_aarray : bigarr }

let memory_create mem_size len =
  { mem_filled = false;
    mem_posn   = 0;
    mem_darray = create_matrix_ba mem_size len 0.;
    mem_uarray = create_matrix_ba mem_size len 0.;
    mem_aarray = make_ba mem_size 0. }

let memory_size mem =
  if mem.mem_filled then length mem.mem_darray else mem.mem_posn

let memory_d mem m =
  if mem.mem_filled
  then
    let len = length mem.mem_darray in
      mem.mem_darray.((len + mem.mem_posn + m - 1) mod len)
  else mem.mem_darray.(m)

let memory_u mem m =
  if mem.mem_filled
  then
    let len = length mem.mem_uarray in
      mem.mem_uarray.((len + mem.mem_posn + m - 1) mod len)
  else mem.mem_uarray.(m)

let memory_a mem m =
  if mem.mem_filled
  then
    let len = length mem.mem_uarray in
      mem.mem_aarray.{(len + mem.mem_posn + m - 1) mod len}
  else mem.mem_aarray.{m}

let memory_add mem d u a =
  let p  = mem.mem_posn in
  let _F = A1.dim d in
    for f = 0 to _F - 1 do
      mem.mem_darray.(p).{f} <- d.{f};
      mem.mem_uarray.(p).{f} <- u.{f};
    done;
    mem.mem_aarray.{mem.mem_posn} <- a;
    if mem.mem_filled
    then mem.mem_posn <- (mem.mem_posn + 1) mod (length mem.mem_darray)
    else (
      if mem.mem_posn == length mem.mem_darray - 1
      then ( mem.mem_posn <- 0; mem.mem_filled <- true; )
      else ( mem.mem_posn <- mem.mem_posn + 1 ) );
    ()

let dot (a1 : bigarr) (a2 : bigarr) =
  let _N = min (A1.dim a1) (A1.dim a2) in
  let rec dr n acc =
    if n >= _N then acc else
      dr (n+1) (acc +. a1.{n} *. a2.{n}) in
    dr 0 0.

let dot_minus _C _F m (a1 : bigarr) (a2 : bigarr) =
  let acc = ref 0. in
    for c = 0 to _C - 1 do
      for f = 0 to _F - 1 do
        let cnf = c * _F + f in
          if cnf < A1.dim a1 && cnf < A1.dim a2 then (
            if f < length m && c < length m.(f) then
              acc := !acc +. a1.{cnf} *. (a2.{cnf} -. m.(f).(c))
            else 
              acc := !acc +. a1.{cnf} *. a2.{cnf}
          );
      done;
    done;
    !acc

let classify_by_wtx (wtx : float array) =
  let _C = length wtx in
  let rec cb mx c =
    if c >= _C then mx
    else if wtx.(c) > wtx.(mx) 
    then cb c  (c+1) 
    else cb mx (c+1) in
    cb 0 1

let compute_error_by_wtx (mass : bigarr) (wtx : float array array) (y : int array) =
  let numerr  = ref 0. in
  let totmass = ref 0. in
  let _N = length wtx in
    if _N == 0 then 0. else (
      let _C = length wtx.(0) in
        if _C > 1 then (
          for n = 0 to _N - 1 do
            if y.(n) != classify_by_wtx wtx.(n) then 
              numerr := !numerr +. if A1.dim mass == 0 then 1. else mass.{n};
            totmass := !totmass +. if A1.dim mass == 0 then 1. else mass.{n};
          done;
          !numerr /. !totmass
        ) else ( 0. ))

let compute_ppx_by_wtx (mass : bigarr) wtx (y : int array) =
  let ppx = ref 0. in
  let _N = length wtx in
  let _C = length wtx.(0) in
    if _C > 1 then (
      for n = 0 to _N - 1 do
        let s = ref zero in
          for c = 0 to _C - 1 do
            s := addLog !s wtx.(n).(c);
          done;
          ppx := !ppx +. wtx.(n).(y.(n)) -. !s;
      done;
      !ppx /. float_of_int _N
    ) else (
      ppx := zero;
      for n = 0 to _N - 1 do
        ppx := !ppx +@ wtx.(n).(0);
      done;
      !ppx -. float_of_int _N
    )

(* (w+e dw)*(w+e dw) = w*w + e^2 dw*dw + 2e w*dw *)

let compute_log_posterior_nonsampled _C lambda y wtx dwtx eta wtw dwtdw dwtw =
  let _N = length wtx in
    (* (w + f dw)*x = w*x + f (dw*x) *)
  let post = ref (0. -. (lambda /. 2.) *. (wtw *. wtw +. eta *. eta *. dwtdw *. dwtdw +. 2. *. eta *. dwtw)) in
    for n = 0 to _N - 1 do
      let sum = ref zero in
        for c = 0 to _C - 1 do
          let wtx' = wtx.(n).(c) +. eta *. dwtx.(n).(c) in
            sum := addLog !sum wtx';
            if c == y.(n) then post := !post +. wtx';
        done;
        post := !post -. !sum;
    done;
    !post

let compute_log_posterior_sampled lambda y wtx dwtx eta wtw dwtdw dwtw samp_expZ =
  let _N = length wtx in
    (* (w + f dw)*x = w*x + f (dw*x) *)
  let post = ref (0. -. (lambda /. 2.) *. (wtw *. wtw +. eta *. eta *. dwtdw *. dwtdw +. 2. *. eta *. dwtw)) in
    for n = 0 to _N - 1 do
      let wtx' = wtx.(n).(0) +. eta *. dwtx.(n).(0) in
        post := !post +. wtx' -. samp_expZ;
    done;
    !post

let compute_log_posterior _C lambda y wtx dwtx eta wtw dwtdw dwtw sampOpt =
  match sampOpt with
      None       -> bad_float_to_neginf (compute_log_posterior_nonsampled _C lambda y wtx dwtx eta wtw dwtdw dwtw)
    | Some (_,z) -> bad_float_to_neginf (compute_log_posterior_sampled       lambda y wtx dwtx eta wtw dwtdw dwtw z)


let perform_line_search _C _F lambda y wtx (w : bigarr) (dw : bigarr) dwtx (g : bigarr) sampOpt mean_model =
  let dwtdw = dot dw dw in 
  let sum    = sqrt dwtdw in
  let scale = if sum <= 100. then 1. else 100. /. sum in

  let dwtdw = dwtdw *. scale *. scale in

  let wtw,dwtw = 
    match mean_model with
        None   -> (dot               w w, dot               dw w *. scale)
      | Some m -> (dot_minus _C _F m w w, dot_minus _C _F m dw w *. scale) in

  let f_old = compute_log_posterior _C lambda y wtx dwtx 0. wtw dwtdw dwtw sampOpt in

  let slope = scale *. dot g dw in
    if slope < 0. then 0. else

  let test = fold_lefti_ba (fun i dwi test ->
                              let temp = (fabs (scale *. dwi)) /. max (fabs w.{i}) 1. in
                                if temp > test then temp else test) 0. dw in
  let alamin = 1e-10 /. test in
  let init_alam = 1. in

  let rec lsrch alam alam2 f2 =
    (* Printf.fprintf stderr "(lsrch %g %g %g) " alam alam2 f2; *)
    if scale *. alam < 1e-6 && is_some sampOpt then 1e-6 *. min 1. scale else
    if scale *. alam < alamin then 1e-8 *. min 1. scale
    else 
      let cur_f = compute_log_posterior _C lambda y wtx dwtx (alam*.scale) wtw dwtdw dwtw sampOpt in
        if f_old <= neg_infinity then 0. (* failwith "" *)
        else if cur_f >= f_old +. 1e-4 *. alam *. scale *. slope then alam *. scale
        else
          if fabs (alam -. init_alam) < 1e-20
          then ( 
            let tmplam = slope /. (2. *. (f_old +. slope -. cur_f)) in
              lsrch (max tmplam (0.1*.alam)) alam cur_f
          ) else (
            let rhs1 = cur_f -. f_old -. alam  *. slope in
            let rhs2 = f2    -. f_old -. alam2 *. slope in
            let a    = (rhs1/.(alam*.alam) -. rhs2/.(alam2*.alam2)) /. (alam-.alam2) in
            let b    = (alam*.rhs2/.(alam2*.alam2) -. alam2*.rhs1/.(alam*.alam)) /. (alam-.alam2) in
            let tmplam = ref 0. in
              if fabs a < 1e-20 then tmplam := slope/.(-2.*.b)
              else (
                let disc = b*.b -. 3.*.a*.slope in
                  if disc < 0. then tmplam := 0.5 *. alam
                  else if b <= 0. then tmplam := ((sqrt disc) -. b) /. (3.*.a)
                  else tmplam := 0. -. slope /. (b +. sqrt disc);
              );
              if !tmplam > 0.5*.alam then tmplam := 0.5*.alam;
              lsrch (max !tmplam (0.1*.alam)) alam cur_f
          ) in
    lsrch init_alam 0. f_old

let compute_gradient_bernoulli_implicit _C lambda (w : bigarr) x y wtx (g : bigarr) =
  let _N = dat_N x in
  let _F = A1.dim w in
  let numF = x.datav_numfeat in
    for n = 0 to _N - 1 do
      let wti = ref zero in
      let _P  = length x.datav_xi.(n) in
      let wt  = dat_wt x n in
        for c' = 0 to _C - 1 do
          wti := addLog !wti wtx.(n).(c');
        done;
	for c = 0 to _C - 1 do
          let cnf = c * numF in
          let factor = (if c == y.(n) then 1. else 0.) -. exp (wtx.(n).(c) -. !wti) in
            g.{cnf} <- g.{cnf} +. factor *. wt;
            for p = 0 to _P - 1 do
              g.{x.datav_xi.(n).(p) + cnf} <- g.{x.datav_xi.(n).(p) + cnf} +. factor *. wt;
            done;
	done;
    done;
    ()
  
let compute_gradient_nonbernoulli_implicit _C lambda (w : bigarr) x y wtx (g : bigarr) =
  let _N = dat_N x in
  let _F = A1.dim w in
  let numF = x.datav_numfeat in
    for n = 0 to _N - 1 do
      let wti = ref zero in
      let _P  = length x.datav_xi.(n) in
      let wt  = dat_wt x n in
        for c' = 0 to _C - 1 do
          wti := addLog !wti wtx.(n).(c')
        done;
	for c = 0 to _C - 1 do
          let cnf = c * numF in
          let factor = (if c == y.(n) then 1. else 0.) -. exp (wtx.(n).(c) -. !wti) in
            g.{cnf} <- g.{cnf} +. factor *. wt;
            for p = 0 to _P - 1 do
              g.{x.datav_xi.(n).(p) + cnf} <- g.{x.datav_xi.(n).(p) + cnf} +. factor *. x.datav_xv.(n).(p) *. wt;
            done;
	done;
    done;
    ()

let compute_gradient_bernoulli_explicit _C lambda (w : bigarr) x y wtx (g : bigarr) =
  let _N = dat_N x in
  let _F = A1.dim w in
    for n = 0 to _N - 1 do
      let wti = ref zero in
      let wt  = dat_wt x n in
        for c' = 0 to _C - 1 do
          wti := addLog !wti wtx.(n).(c');
        done;
	for c = 0 to _C - 1 do
          let _P  = length x.datav_xi_c.(n).(c) in
          let factor = (if c == y.(n) then 1. else 0.) -. exp (wtx.(n).(c) -. !wti) in
            for p = 0 to _P - 1 do
              g.{x.datav_xi_c.(n).(c).(p)} <- g.{x.datav_xi_c.(n).(c).(p)} +. factor *. wt;
            done;
	done;
    done;
    ()
  
let compute_gradient_nonbernoulli_explicit _C lambda (w : bigarr) x y wtx (g : bigarr) =
  let _N = dat_N x in
  let _F = A1.dim w in
    for n = 0 to _N - 1 do
      let wti = ref zero in
      let wt  = dat_wt x n in
        for c' = 0 to _C - 1 do
          wti := addLog !wti wtx.(n).(c')
        done;
	for c = 0 to _C - 1 do
          let _P  = length x.datav_xi_c.(n).(c) in
          let factor = 
            (if c == y.(n) then 1. else 0.) -. exp (wtx.(n).(c) -. !wti) in
            for p = 0 to _P - 1 do
              g.{x.datav_xi_c.(n).(c).(p)} <- g.{x.datav_xi_c.(n).(c).(p)} +. factor *. x.datav_xv_c.(n).(c).(p) *. wt;
            done;
	done;
    done;
    ()

let compute_gradient_sampled lambda (w : bigarr) x y wtx (g : bigarr) (samp_exp_F : bigarr) samp_exp_Z =
  let _N = dat_N x in
  let _F = A1.dim w in
  let numF = x.datav_numfeat in
    for f = 0 to _F - 1 do
      g.{f} <- g.{f} -. (float_of_int _N) *. (if f == 0 then 1. else samp_exp_F.{f});
    done;
    for n = 0 to _N - 1 do
      let wti = ref zero in
      let _P  = length x.datav_xi.(n) in
      let wt  = dat_wt x n in
        if x.datav_implicit then (
          for p = 0 to _P - 1 do
            g.{x.datav_xi.(n).(p)} <- g.{x.datav_xi.(n).(p)} +. wt;
          done;
        ) else (
          for p = 0 to _P - 1 do
            g.{x.datav_xi.(n).(p)} <- g.{x.datav_xi.(n).(p)} +. wt *. x.datav_xv.(n).(p);
          done;
        );
    done;
    ()

let compute_gradient _C lambda (w : bigarr) x y wtx (g : bigarr) (sampOption : (bigarr * float) option) mean_model =
  let _F = A1.dim w in
    (match mean_model with
         None -> (
           for f = 0 to _F - 1 do 
             g.{f} <- 0. -. lambda *. w.{f}; 
           done)
       | Some m -> (
           for c = 0 to _C - 1 do
             for f = 0 to x.datav_numfeat - 1 do
               let cnf = c * x.datav_numfeat + f in
                 if f < length m && c < length m.(f) then
                   g.{cnf} <- 0. -. lambda *. (w.{cnf} -. m.(f).(c))
                 else 
                   g.{cnf} <- 0. -. lambda *. w.{cnf};
             done;
           done)
    );
    match sampOption with
        None ->
          if x.datav_implicit
          then
            if x.datav_bernoulli 
            then compute_gradient_bernoulli_implicit    _C lambda w x y wtx g
            else compute_gradient_nonbernoulli_implicit _C lambda w x y wtx g
          else
            if x.datav_bernoulli
            then compute_gradient_bernoulli_explicit    _C lambda w x y wtx g
            else compute_gradient_nonbernoulli_explicit _C lambda w x y wtx g
      | Some (f,z) ->
          compute_gradient_sampled lambda w x y wtx g f z

  

(* compute_gradient is where we spend all our time -- let's optimize it
   for the different data types separately...original version is commented
   below, and takes:
     44.37%=111.44sec in 41 calls, 2.72/2.76 sec/call    --  without prep
     16.33%=108.47sec in 41 calls, 0.83/1.70 sec/call    --  with prep

   when we optimize for bernoulli/implicit, this drops to
     18.92%=80.97sec  in 41 calls, 0.67/0.67 sec/call

let compute_gradient _C lambda w x y wtx g =
  let _N = dat_N x in
  let _F = length w in
  let _  = for f = 0 to _F - 1 do g.(f) <- 0. -. lambda *. w.(f); done in

  let add_to_grad w n c =
    let _P = dat_P x n c in
      for p = 0 to _P - 1 do
	g.(dat_i x n c p) <- g.(dat_i x n c p) +. w *. dat_v x n c p;
      done in

    for n = 0 to _N - 1 do
      let wti = ref zero in
        for c' = 0 to _C - 1 do
          wti := addLog !wti wtx.(n).(c')
        done;
	for c = 0 to _C - 1 do
          if c == y.(n)
          then add_to_grad (1. -. exp (wtx.(n).(c) -. !wti)) n c
          else add_to_grad (0. -. exp (wtx.(n).(c) -. !wti)) n c;
	done;
    done;
    ()
*)

let compute_update_direction_bfgs _C lambda memory x y wtx (w : bigarr) (w_old : bigarr) (g_old : bigarr) (q : bigarr) (g : bigarr) sampOption mean_model =
  let _N = dat_N x in
  let _F = A1.dim w in
  (* first, compute the gradient *)
  let () = compute_gradient _C lambda w x y wtx g sampOption mean_model in

  (* now, compute u, d, alpha and sigma *)
  let utd = ref 0. in let utu = ref 0. in
    for f = 0 to _F - 1 do
      let u = g.{f} -. g_old.{f} in
      let d = w.{f} -. w_old.{f} in
	utd := !utd +. u*.d;
	utu := !utu +. u*.u;
	g_old.{f} <- u;
	w_old.{f} <- d;
    done;

  let alpha = !utd in 
  let sigma = !utd /. !utu in

  (* update our memory *)
    memory_add memory w_old g_old alpha;

  (* now, compute q = Hg *)
  let _    = for f = 0 to _F - 1 do q.{f} <- g.{f}; done in
  let _M   = memory_size memory in
  let beta = make _M 0. in
    for m = _M - 1 downto 0 do
      beta.(m) <- (dot (memory_d memory m) g) /. memory_a memory m;
      let um = memory_u memory m in
        for f = 0 to _F - 1 do
          q.{f} <- q.{f} -. beta.(m) *. um.{f};
        done;
    done;
    for f = 0 to _F - 1 do
      q.{f} <- q.{f} *. sigma;
    done;
    for m = 0 to _M - 1 do
      let dm  = memory_d memory m in
      let um  = memory_u memory m in
      let am  = memory_a memory m in
      let umq = ref (dot um q) in
        for f = 0 to _F - 1 do
          let dq = dm.{f} *. (beta.(m) -. !umq /. am) in
            umq := !umq +. um.{f} *. dq;
            q.{f} <- q.{f} +. dq;
        done;
    done;

  (* replace g_old with g *)
    for f = 0 to _F - 1 do
      g_old.{f} <- g.{f} -. g_old.{f};
    done;

    ()

let update_wtx_bfgs _C x wtx eta dwtx =
  let _N = dat_N x  in
    for n = 0 to _N - 1 do
      for c = 0 to _C - 1 do
	wtx.(n).(c) <- wtx.(n).(c) +. bad_float_to_zero (eta *. dwtx.(n).(c));
      done;
    done;
    ()

let update_wtx_nodwtx _C x wtx eta (dw : bigarr) =
  let _N = dat_N x in
    for n = 0 to _N - 1 do
      for c = 0 to _C - 1 do
        let _P = dat_P x n c in
          if x.datav_implicit then
            wtx.(n).(c) <- wtx.(n).(c) +. bad_float_to_zero (eta *. dw.{c * x.datav_numfeat});
          for p = 0 to _P - 1 do
(*            Printf.eprintf "(%d) " (dat_i x n c p); flush stderr;
            Printf.eprintf "{%g} " (dat_v x n c p); flush stderr;
            Printf.eprintf "[%g] " dw.{dat_i x n c p}; flush stderr; *)
            wtx.(n).(c) <- wtx.(n).(c) +. bad_float_to_zero (eta *. dw.{dat_i x n c p} *. dat_v x n c p);
          done;
      done;
    done;
    ()

(* a lot of time is spent in update_weights after compute_wtx has been
   inlined, so we also want to optimize this for different data types.  
   the original version is included below:
    for bernoulli, implicit, this drops update_weights from
      36%, 74.63sec, 41 calls at 1.82/2.43 s/call
    to
      20%, 50.13sec, 41 calls at 0.57/1.19 s/call
      
let compute_wtx _C x dw wtx =
  let _N = dat_N x in
    for n = 0 to _N - 1 do
      for c = 0 to _C - 1 do
	let _P = dat_P x n c in
	  wtx.(n).(c) <- 0.;
	  for p = 0 to _P - 1 do
	    wtx.(n).(c) <- wtx.(n).(c) +. dw.(dat_i x n c p) *. dat_v x n c p;
	  done;
      done;
    done;
    ()
*)

let compute_wtx_bernoulli_implicit _C x (dw : bigarr) wtx =
  let _N = dat_N x in
  let _F = x.datav_numfeat in
    for n = 0 to _N - 1 do
      let _P = length x.datav_xi.(n) in
        for c = 0 to _C - 1 do
          let cnf = c * _F in
          wtx.(n).(c) <- dw.{cnf};
          for p = 0 to _P - 1 do
            wtx.(n).(c) <- wtx.(n).(c) +. dw.{x.datav_xi.(n).(p) + cnf};
          done;
        done;
    done;
    ()

let compute_wtx_nonbernoulli_implicit _C x (dw : bigarr) wtx =
  let _N = dat_N x in
  let _F = x.datav_numfeat in
    for n = 0 to _N - 1 do
      let _P = length x.datav_xi.(n) in
        for c = 0 to _C - 1 do
          let cnf = _F * c in
          wtx.(n).(c) <- dw.{cnf};
          for p = 0 to _P - 1 do
            wtx.(n).(c) <- wtx.(n).(c) +. dw.{x.datav_xi.(n).(p)+cnf} *. x.datav_xv.(n).(p);
          done;
        done;
    done;
    ()

let compute_wtx_bernoulli_explicit _C x (dw : bigarr) wtx =
  let _N = dat_N x in
    for n = 0 to _N - 1 do
        for c = 0 to _C - 1 do
          let _P = length x.datav_xi_c.(n).(c) in
          wtx.(n).(c) <- dw.{0};
          for p = 0 to _P - 1 do
            wtx.(n).(c) <- wtx.(n).(c) +. dw.{x.datav_xi_c.(n).(c).(p)};
          done;
        done;
    done;
    ()

let compute_wtx_nonbernoulli_explicit _C x (dw : bigarr) wtx =
  let _N = dat_N x in
    for n = 0 to _N - 1 do
        for c = 0 to _C - 1 do
          let _P = length x.datav_xi_c.(n).(c) in
          wtx.(n).(c) <- dw.{0};
          for p = 0 to _P - 1 do
            wtx.(n).(c) <- wtx.(n).(c) +. dw.{x.datav_xi_c.(n).(c).(p)} *. x.datav_xv_c.(n).(c).(p);
          done;
        done;
    done;
    ()

let update_weights _C lambda x y wtx (w : bigarr) (w_old : bigarr) (dw : bigarr) dwtx (g : bigarr) xDe wDe xTe wTe sampOpt mean_model =
  let _   = 
    if x.datav_implicit
    then
      if x.datav_bernoulli 
      then compute_wtx_bernoulli_implicit    _C x dw dwtx
      else compute_wtx_nonbernoulli_implicit _C x dw dwtx
    else
      if x.datav_bernoulli 
      then compute_wtx_bernoulli_explicit    _C x dw dwtx
      else compute_wtx_nonbernoulli_explicit _C x dw dwtx in
  let eta = perform_line_search _C x.datav_numfeat lambda y wtx w dw dwtx g sampOpt mean_model in
  let _N = dat_N x in
  let _F = A1.dim w in
    (* update wtx *)
    update_wtx_bfgs   _C x   wtx eta dwtx;
    update_wtx_nodwtx _C xDe wDe eta dw;
    update_wtx_nodwtx _C xTe wTe eta dw;

    (* update w and w_old *)
    for f = 0 to _F - 1 do
      w_old.{f} <- w.{f};
      w.{f}     <- w.{f} +. bad_float_to_zero (eta *. dw.{f});
    done;

    (* return amount of change *)
    eta *. fold_left_ba (fun mx v -> max mx (fabs v)) 0. dw

let z_only = function None -> None | Some (_,z) -> Some z

let optimize_bfgs quiet maxi lambda dpp _M x y xDe yDe xTe yTe (w : bigarr) wtx wtxDev wtxTst sampOption mean_model returnStuff =
  let _N = dat_N x in
  let _C = x.datav_numclass in
  let _F = dat_F x in

  let _ = if not quiet then printIter 1 maxi in

  let w_old  = make_ba _F 0. in
  let g      = make_ba _F 0. in
  let g_old  = make_ba _F 0. in

  let do_samp w =
    match sampOption with
        None -> None
      | Some ((prog,file),args,(burn,num,space)) -> 
          if prog = "" && file = "" then failwith "optimize_bfgs: sample estimation requires either program or file"
          else if prog <> "" && file <> "" then failwith "optimize_bfgs: sample estimation requires only one of program and file"
          else if prog <> ""
          then Some (Wsemlm.independence_sampling _F 
                       (Wsemlm.mk_external_gen_q       prog args (fun s -> get_vocab s) 
                                                       (1 + burn + num*space) x.datav_implicit)
                       w burn num space)
          else Some (Wsemlm.independence_sampling _F
                       (Wsemlm.mk_gen_q_from_file_list file      (fun s -> get_vocab s) 
                                                       (1 + burn + num*space) x.datav_implicit)
                       w burn num space) in

  let samp0 = do_samp w_old in

  let ()     = compute_gradient _C lambda w x y wtx g samp0 mean_model in
  let dwtx   = create_matrix _N _C 0. in
  let gtg    = sqrt (dot g g) in
  let step   = make_ba _F 0. in
  let _      = iteri_ba (fun i v -> step.{i} <- v /. gtg) g in
  let myfstc = update_weights _C lambda x y wtx w w_old step dwtx g xDe wtxDev xTe wtxTst samp0 mean_model in

  let _  = if not quiet then printDW myfstc in
  let pp,_ = print_info quiet compute_ppx_by_wtx compute_error_by_wtx x.datav_weights wtx y xDe.datav_weights wtxDev yDe xTe.datav_weights wtxTst yTe in 
  let _  = if not quiet then (Printf.fprintf stderr "\n"; flush stderr) in

  let _ = for f = 0 to _F - 1 do g_old.{f} <- g.{f}; done in

  let memory = memory_create _M _F in

  let rec iterate lambda inum lastpp =
    if not quiet then printIter inum maxi;
    let samp = do_samp w in
    let () = compute_update_direction_bfgs _C lambda memory x y wtx w w_old g_old step g samp mean_model in
      for f = 0 to _F - 1 do 
        step.{f} <- bad_float_to_zero (0. -. step.{f}); 
      done;
    let dw = update_weights _C lambda x y wtx w w_old step dwtx g  xDe wtxDev xTe wtxTst samp mean_model in
    if not quiet then printDW dw;
    let pp,de = print_info quiet compute_ppx_by_wtx compute_error_by_wtx x.datav_weights wtx y xDe.datav_weights wtxDev yDe xTe.datav_weights wtxTst yTe in 
    if not quiet then (Printf.fprintf stderr "\n"; flush stderr);
      
    for f = 0 to _F - 1 do
      g_old.{f} <- g.{f};
    done;

    if inum < maxi && fabs(lastpp-.pp) > dpp && dw > 1e-20
    then iterate lambda (inum+1) pp 
    else (de,z_only samp,
          if returnStuff
          then Some wtx
          else None )in 

  iterate lambda 2 myfstc


let optimize quiet maxi optimize_l lambda dpp _M x y xDe yDe xTe yTe sampOption mean_model init_model returnStuff =
  let _N = dat_N x in
  let _C = x.datav_numclass in
  let _F = dat_F x in

  let w      = make_ba _F 0. in
  let wtx    = create_matrix _N _C 0. in
  let wtxDev = create_matrix (dat_N xDe) _C 0. in
  let wtxTst = create_matrix (dat_N xTe) _C 0. in

  let or_empty x n = if n >= length x then [||] else x.(n) in

  let _ = match init_model with None -> () | Some wi -> (
    for f = 0 to min (x.datav_numfeat-1) (length wi-1) do
      for c = 0 to min (length wi.(f)-1) (_C-1) do
        w.{f + c * x.datav_numfeat} <- wi.(f).(c);
      done;
    done;
    for c = 0 to _C-1 do
      for n = 0 to dat_N x   - 1 do wtx.(   n).(c) <- dotEx wi c   x.datav_xi.(n) (or_empty   x.datav_xv n); done;
      for n = 0 to dat_N xDe - 1 do wtxDev.(n).(c) <- dotEx wi c xDe.datav_xi.(n) (or_empty xDe.datav_xv n); done;
      for n = 0 to dat_N xTe - 1 do wtxTst.(n).(c) <- dotEx wi c xTe.datav_xi.(n) (or_empty xTe.datav_xv n); done;
    done) in

  let best_w = ref None in let best_w_de = ref 0. in
  let best_stuff = ref None in

  let rec opt_lambda lambda =
    let _ = Printf.fprintf stderr "optimizing with lambda = %g\n" lambda in
    let de,_,stuff = optimize_bfgs quiet maxi lambda dpp _M x y xDe yDe xTe yTe w wtx wtxDev wtxTst sampOption mean_model returnStuff in
      if is_some !best_w && lambda > 0. && de -. !best_w_de > -2e-6
      then from_some !best_w, None, !best_stuff
      else ( best_w := Some (copy_ba w) ; best_w_de := de ; best_stuff := stuff ;
             if de > 0. then opt_lambda (lambda *. 0.5) else w,None,stuff ) in
    
    if optimize_l 
    then opt_lambda lambda
    else 
      let _,samp,stuff = optimize_bfgs quiet maxi lambda dpp _M x y xDe yDe xTe yTe w wtx wtxDev wtxTst sampOption mean_model returnStuff in 
        w,samp,stuff


