(* 
for debugging/safe execution, compile with:
  ocamlc -g str.cma util.ml data.ml cg.ml bfgs.ml main.ml -o megam
for super fast, perhaps core-dumping execution, compile with:
  ocamlopt -ffast-math -inline 9999 -unsafe str.cmxa util.ml data.ml cg.ml bfgs.ml main.ml -o megam
*)

open Sys
open Array
open Util
open Data

let usage_text = "usage: megam [options] <model-type> <input-file>
[options] are any of:
   -fvals           data is in not in bernoulli format (i.e. feature
                    values appear next to their features int the file)

   -explicit        this specifies that each class gets its own feature
                    vector explicitely, rather than using the same one
                    independently of class
                    (only valid for multiclass problems)

   -quiet           don't generate per-iteration output

   -maxi <int>      specify the maximum number of iterations
                    (default: 100)

   -dpp <float>     specify the minimum change in perplexity
                    (default: -99999)

   -memory <int>    specify the memory size for LM-BFGS (multiclass only)
                    (default: 5)

   -lambda <float>  specify the precision of the Gaussian prior
                    (default: 1)

   -tune            tune lambda using repeated optimizations (starts with
                    specified -lambda value and drops by half each time
                    until optimal dev error rate is achieved)

   -sprog <prog>    for density estimation problems, specify the program 
                    that will generate samples for us (see also -sfile)

   -sfile <files>   for de problems, instead of -sprog, just read from a
                    (set of) file(s); specify as file1:file2:...:fileN

   -sargs <string>  set the arguments for -sprog; default \"\"

   -sspec <i,i,i>   set the <burn-in time, number of samples, sample space>
                    parameters; default: 1000,500,50

   -sforcef <file>  include features listed in <file> in feature vectors
                    (even if they don't exist in the training data)

   -predict <file>  load parameters from <file> and do prediction
                    (will not optimize a model)

   -mean <file>     the Gaussian prior typically assumes mu=0 for all features;
                    you can instead list means in <file> in the same format
                    as is output by this program (baseline adaptation)

   -init <file>     initialized weights as in <file>

   -abffs <int>     use approximate Bayes factor feature selection; add features
                    in batches of (at most) <int> size

<model-type> is one of:
   binary           this is a binary classification problem; classes
                    are determined at a threshold of 0.5 (anything
                    less is negative class, anything greater is positive)

   alma             binary classification with large margin updates

   binomial         this is a binomial problem; all values should be
                    in the range [0,1]

   multiclass       this is a multiclass problem; classes should be
                    numbered [0, 1, 2, ...]; anything < 0 is mapped
                    to class 0

   density          this is a density estimation problem and thus the
                    partition function must be calculated through samples
                    (must use -sprog or -sfile arguments, above)
"

let arg_error str = 
  Printf.fprintf stderr "%s\n" usage_text;
  failwith ("Error: " ^ str)

let l = length argv
let _ = if l < 3 then arg_error "not enough arguments"
let ber = ref true
let imp = ref true
let maxi = ref 100
let dpp  = ref (-99999.)
let memory = ref 5
let lambda = ref 1.
let pred_file = ref None
let mean_file = ref None
let quiet = ref false
let init_file = ref None
let abffs = ref None
let sprog = ref ""
let sfile = ref ""
let sargs = ref ""
let sspec = ref (100,500,50)
let sforcef = ref ""
let tune = ref false
let opt = ref
  (match argv.(l-2) with
       "binary"     -> Binary
     | "alma"       -> Alma
     | "binomial"   -> Binomial
     | "multiclass" -> Multiclass
     | "density"    -> Sampled
     | s -> arg_error ("'" ^ s ^ "' is not a valid optimization type"))
let fname = argv.(l-1)

let read_sspec s =
  match Str.split (Str.regexp ",") s with
      [a;b;c] -> int_of_string a, int_of_string b, int_of_string c
    | _ -> failwith ""

let _ = 
  let rec read_opt p =
    if p >= l-2 then () else
      match argv.(p) with
          "-fvals"    -> ( ber   := false ; read_opt (p+1) )
        | "-explicit" -> ( imp   := false ; read_opt (p+1) )
        | "-tune"     -> ( tune  := true  ; read_opt (p+1) )
        | "-quiet"    -> ( quiet := true  ; read_opt (p+1) )
        | "-maxi"     -> ( (try maxi := int_of_string argv.(p+1) with _ -> arg_error "-maxi expects an integer argument")
                          ; read_opt (p+2) )
        | "-dpp"      -> ( (try dpp  := float_of_string argv.(p+1) with _ -> arg_error "-dpp expects a float argument")
                          ; read_opt (p+2) )
        | "-memory"   -> ( (try memory := int_of_string argv.(p+1) with _ -> arg_error "-memory expects an integer argument")
                          ; read_opt (p+2) )
        | "-lambda"   -> ( (try lambda := float_of_string argv.(p+1) with _ -> arg_error "-lambda expects a float argument")
                          ; read_opt (p+2) )
        | "-sprog"    -> ( (try sprog := argv.(p+1) with _ -> arg_error "-sprog expects an argument")
                          ; read_opt (p+2) )
        | "-sfile"    -> ( (try sfile := argv.(p+1) ; sforcef := argv.(p+1) with _ -> arg_error "-sfile expects an argument")
                          ; read_opt (p+2) )
        | "-sargs"    -> ( (try sargs := argv.(p+1) with _ -> arg_error "-sargs expects an argument")
                          ; read_opt (p+2) )
        | "-sspec"    -> ( (try sspec := read_sspec argv.(p+1) with _ -> arg_error "-sspec expects an argument of the form 'burnin,numsamples,spacing'")
                          ; read_opt (p+2) )
        | "-sforcef"  -> ( (try sforcef := argv.(p+1) with _ -> arg_error "-sforcef expects an argument")
                          ; read_opt (p+2) )
        | "-predict"  -> ( (try pred_file := Some argv.(p+1) with _ -> arg_error "-predict expects a file argument")
                          ; read_opt (p+2) )
        | "-mean"     -> ( (try mean_file := Some argv.(p+1) with _ -> arg_error "-mean expects a file argument")
                          ; read_opt (p+2) )
        | "-init"     -> ( (try init_file := Some argv.(p+1) with _ -> arg_error "-init expects a file argument")
                          ; read_opt (p+2) )
        | "-abffs"    -> ( (try abffs := Some (int_of_string argv.(p+1)) with _ -> arg_error "-abffs expects an integer argument")
                          ; read_opt (p+2) )
        | _ -> arg_error ("unknown option: '" ^ argv.(p) ^ "'") in
    read_opt 1

(* now, check for stuff that doesn't make sense *)

let _ = if !maxi   < 1  then arg_error "-maxi should be >= 1"
let _ = if !memory < 2  then arg_error "-memory should be >= 2"
let _ = if !lambda < 0. then arg_error "-lambda should be >= 0"

let _ = if !opt == Sampled && !sprog = "" && !sfile = "" && !pred_file = None then 
          arg_error "need to specify -sprog for sampled problems"
let _ = if !opt != Sampled && !sprog <> "" then arg_error "spurious use of -sprog for non-sampled problem"
let _ = if !opt != Sampled && !sfile <> "" then arg_error "spurious use of -sfile for non-sampled problem"
let _ = if !sfile <> "" && !sprog <> "" then
          arg_error "only one of -sprog and -sfile allowed"

let _ = if not !imp && !opt = Binomial then arg_error "cannot use -explicit with binomial optimization"
let _ = if not !imp && !opt = Binary then (
  Printf.fprintf stderr "Warning: it makes no sense to use -explicit with binary optimization\n";
  Printf.fprintf stderr "         we will proceed as if you had said multiclass, but the output\n";
  Printf.fprintf stderr "         will differ slightly from standard binary weights\n";
  opt := Multiclass )

let _ = if is_some !abffs && (not !imp (* || not !ber *) ) then arg_error "can only use feature selection with implicit data"
let _ = if is_some !abffs && !opt != Multiclass then arg_error "can only use feature selection with multiclass problems"

let model = 
  match !pred_file with
      None -> None
    | Some f -> (
        let model_h = open_in f in
        let w = read_model model_h !opt in
        let _ = close_in model_h in
          Some w)

let init_model = 
  match !init_file with
      None -> None
    | Some f -> (
        let model_h = open_in f in
        let w = read_model model_h !opt in
        let _ = close_in model_h in
          Some w)

let mean_model = 
  match !mean_file with
      None -> None
    | Some f -> (
        let model_h = open_in f in
        let w = read_model model_h !opt in
        let _ = close_in model_h in
          Some w)

let pred_err = ref 0.
let pred_cnt = ref 0

let warnedMissingZ = ref false

let predict_imp y x v =
  match model with
      None -> ()
    | Some w ->
        (match !opt with
             Binary -> 
               let p = (1. /. (1. +. exp (0. -. dotEx w 0 x v))) in
               let _ = incr pred_cnt in
               let _ = if y > 0.5 && p < 0.5 || y < 0.5 && p > 0.5 then pred_err := !pred_err +. 1. in
                 Printf.printf "%d\t%10.20f\n" (if p < 0.5 then 0 else 1) p
           | Binomial ->
               let p = (1. /. (1. +. exp (0. -. dotEx w 0 x v))) in
               let _ = incr pred_cnt in
               let _ = pred_err := !pred_err +. (p -. y) *. (p -. y) in
                 Printf.printf "%d\t%10.20f\n" (if p < 0.5 then 0 else 1) p
           | Multiclass -> failwith "predict_imp (Multiclass): shouldn't get here"
	   | Sampled -> 
               let _Z = 
                 try  
                   let _Zid = Hashtbl.find vocabFwd "***NORMALIZING-CONSTANT-Z***" in
                     w.(_Zid).(0)
                 with Not_found -> (
                   if not !warnedMissingZ
                   then ( Printf.fprintf stderr "WARNING: Normalizing constant not found -- not normalizing\n";
                          warnedMissingZ := true );
                   one ) in
               let p    = ((w.(0).(0) +. dotEx w 0 x v) *@ _Z) in
               let _    = incr pred_cnt in
                 Printf.printf "%10.20f\n" p
        )

let predict_imp_mc y x v =
  match model with
      None -> ()
    | Some w ->
        (match !opt with
             Binary | Binomial -> failwith "shouldn't get here"
           | Multiclass -> (
               let _C = length w.(0) in
               let p  = init _C (fun c -> dotEx w c x v) in
               let s  = fold_left addLog zero p in
               let _  = iteri (fun i v -> p.(i) <- exp (v -. s)) p in
               let mxp = ref 0 in
                 for c = 1 to _C - 1 do
                   if p.(c) > p.(!mxp) then mxp := c;
                 done;
                 Printf.printf "%d\t" !mxp;
                 for c = 0 to _C - 1 do
                   if c > 0 then Printf.printf " ";
                   Printf.printf "%10.20f" p.(c);
                 done;
                 incr pred_cnt;
                 if !mxp != y then pred_err := !pred_err +. 1.;
                 Printf.printf "\n"
             )
	   | Sampled -> failwith "predict_imp_mc (Sampled): not yet implemented"
	)

let predict_exp y x v =
  match model with
      None -> ()
    | Some w ->
        (match !opt with
             Binary | Binomial | Sampled -> failwith "predict_exp: shouldn't get here"
           | Multiclass -> (
               let _C = length x in
               let p  = init _C (fun c -> dotEx w c x.(c) v.(c)) in
               let s  = fold_left addLog zero p in
               let _  = iteri (fun i v -> p.(i) <- exp (v -. s)) p in
               let mxp = ref 0 in
                 for c = 1 to _C - 1 do
                   if p.(c) > p.(!mxp) then mxp := c;
                 done;
                 Printf.printf "%d\t" !mxp;
                 for c = 0 to _C - 1 do
                   if c > 0 then Printf.printf " ";
                   Printf.printf "%10.20f" p.(c);
                 done;
                 incr pred_cnt;
                 if !mxp != y then pred_err := !pred_err +. 1.;
                 Printf.printf "\n"
             ))

(* silly functions to read classes *)
let read_binary_class ln str =
  try
    let i = float_of_string str in
      if i < 0.5 then -1. else 1.
  with Failure err ->
    failwith ("error: \"" ^ err ^ "\" occured on line " ^ string_of_int ln ^ " when trying to parse \"" ^ str ^ "\" as a float for binary class number")
  
let read_binomial_class ln str =
  try
    let i = float_of_string str in
      if i < 0. then 0. else if i > 1. then 1. else i
  with Failure err ->
    failwith ("error: \"" ^ err ^ "\" occured on line " ^ string_of_int ln ^ " when trying to parse \"" ^ str ^ "\" as a float for binomial class value")

let max_read_class = ref 0
let read_multiclass ln str =
  try
    let i = int_of_string str in
    let j = if i < 0 then 0 else i in
      max_read_class := max !max_read_class j;
      j
  with Failure err ->
    failwith ("error: \"" ^ err ^ "\" occured on line " ^ string_of_int ln ^ " when trying to parse \"" ^ str ^ "\" as an int for multiclass class")


let map_flipped list f = List.map f list

(* read forced features *)
let read_forced_features () =
  if !sforcef = "" then () else (
    ignore (map_flipped (Str.split (Str.regexp ":") !sforcef)
      (fun fname ->
         let h = open_in fname in
         let rec rd () =
           match maybe_read_line h with
	       None -> ()
	     | Some l -> (
	         List.iter (fun s -> ignore (get_vocab s)) 
	                    (Str.split (Str.regexp "[ \t]+") l);
	         rd ()) in
           rd ();
           close_in h;
           ()));
    ();
  )

(* now, do the optimization specified *)

let h  = open_in fname
let () =
  match !opt with
      Binary     -> ( (* must be implicit *)
        let (xTr,yTr),(xDe,yDe),(xTe,yTe) = load_implicit !quiet !ber h read_binary_class 1. false predict_imp in
        let _ = close_in h in
          if !pred_file == None then (
            let w = Cg.compute !quiet !maxi  !tune !lambda !dpp xTr yTr xDe yDe xTe yTe mean_model init_model in
              printVectorVocab w ))

    | Alma -> (
        let (xTr,yTr),(xDe,yDe),(xTe,yTe) = load_implicit !quiet !ber h read_binary_class 1. false predict_imp in
        let _ = close_in h in
          if !pred_file == None then (
            let w = Alma.optimize !quiet !maxi !lambda xTr yTr xDe yDe xTe yTe None (Some (xDe,yDe)) in
              printVectorVocab_ba w ))

    | Binomial   -> ( (* must be implicit *)
        let (xTr,yTr),(xDe,yDe),(xTe,yTe) = load_implicit !quiet !ber h read_binomial_class 1. false predict_imp in
        let _ = close_in h in
          if !pred_file == None then (
            let w = Cg.computeBinomial !quiet !maxi !tune !lambda !dpp xTr yTr xDe yDe xTe yTe mean_model init_model in
              printVectorVocab w ))

    | Sampled    -> (* must be implicit *)
        let (xTr,yTr),(xDe,yDe),(xTe,yTe) = load_implicit !quiet !ber h read_binomial_class 1. false predict_imp in
          close_in h;
          if !pred_file = None then (
	    read_forced_features ();
	    xTr.datav_numfeat <- 1 + !vocabSize;
            let w,s_z,_ = Bfgs.optimize !quiet !maxi !tune !lambda !dpp !memory xTr [||] xDe [||] xTe [||] (Some ((!sprog, !sfile), !sargs, !sspec)) mean_model init_model false in
              (match s_z with None -> () | Some z -> Printf.printf "***NORMALIZING-CONSTANT-Z***\t%10.20f\n" z);
              printVectorVocab_ba w
          );

    | Multiclass -> ( (* might or might not be implicit *)
        let (xTr,yTr),(xDe,yDe),(xTe,yTe) = 
          if !imp
          then load_implicit !quiet !ber h read_multiclass 0 false predict_imp_mc
          else load_explicit !quiet !ber h predict_exp in
          if !pred_file == None then (
            let _ = if !imp then (
              xTr.datav_numclass <- 1 + !max_read_class;
              xDe.datav_numclass <- 1 + !max_read_class;
              xTe.datav_numclass <- 1 + !max_read_class ) in
            let _ =
              if !max_read_class == 1 && is_none !abffs then (
                Printf.fprintf stderr "Warning: there only appear to be two classes, but we're\n";
                Printf.fprintf stderr "         optimizing with BFGS...using binary optimization\n";
                Printf.fprintf stderr "         with CG would be much faster\n" ) in

            let w =
              match !abffs with
                  None ->
                    let w,_,_ = Bfgs.optimize !quiet !maxi !tune !lambda !dpp !memory xTr yTr xDe yDe xTe yTe None mean_model init_model false in
                      w
                | Some _S ->
                    Abffs.abffs_selection !quiet !maxi !tune !lambda !dpp !memory xTr yTr xDe yDe xTe yTe None mean_model init_model _S in

              if not !imp then printVectorVocab_ba w else (
                for f = 0 to xTr.datav_numfeat do
	          let offset = w.{f} in
                  let keep = ref false in
                    for c = 1 to xTr.datav_numclass - 1 do 
                      if abs_float (w.{f + c * xTr.datav_numfeat} -. offset) > 1e-10 then keep := true;
                    done;
                    if !keep then (
	              if f == 0 then Printf.printf "**BIAS**\t0" else Printf.printf "%s\t0" (Hashtbl.find vocabBwd f);
	              for c = 1 to xTr.datav_numclass - 1 do
	                Printf.printf " %10.20f" ((w.{f + c * xTr.datav_numfeat} -. offset));
	              done;
	              Printf.printf "\n";
                    );
	            flush stdout;
                done
              ))
      )

let _ =
  match !pred_file with None -> () | Some _ ->
    Printf.fprintf stderr "Error rate = %g / %g = %g\n" 
      !pred_err (float_of_int !pred_cnt) (!pred_err /. float_of_int !pred_cnt)
